#include <kernel.h>
#include <task.h>
#include <vfs.h>
#include <mm.h>

#include <arch/descriptor.h>
#include <arch/vga.h>
#include <arch/io.h>
#include <arch/pc_keyb.h>
#include <arch/timer.h>
#include <arch/8259a.h>
#include <arch/idt.h>

#include <arch/multiboot.h>

/* Global descriptor table */
dt_entry_t __desc_aligned GDT[]={
	{dummy:0},
	
	/* Kernel space */
	stnd_desc(0, 0xFFFFF, (D_CODE | D_READ  | D_BIG | D_BIG_LIM)),
	stnd_desc(0, 0xFFFFF, (D_DATA | D_WRITE | D_BIG | D_BIG_LIM)),
	
	/* User space */
	stnd_desc(0, 0xFFFFF, (D_CODE | D_READ  | D_BIG | D_BIG_LIM | D_DPL3)),
	stnd_desc(0, 0xFFFFF, (D_DATA | D_WRITE | D_BIG | D_BIG_LIM | D_DPL3)),
};
struct gdtr loadgdt={ sizeof(GDT)-1, (u_int32_t)GDT};

void idle_task(void);

/* beginning and end of kernel */
extern u_int32_t _begin;
extern u_int32_t _end;
extern u_int32_t _etext;
extern u_int32_t __init_start;

/* For bootmem allocator */
u_int32_t __bootmem_end=0x4000;

/* lower and upper memory */
u_int32_t mem_lo, mem_hi;

/* Bootup CPU info */
struct cpu_info cpu_bsp;

char *cmdline;

char *cpuid_str[]={
	"fpu",  "vme",   "de",   "pse",
	"tse",  "msr2",  "pae",  "mce",
	"cx8",  "apic",  "_r1",  "sep",
	"mtrr", "pge",   "mca",  "cmov",
	"pat",  "pse36", "psn",  "clflush",
	"_r2",  "dts",   "acpi", "mmx",
	"fxsr", "sse",   "sse2", "ss",
	"htt",  "tm",    "ia64",  "pbe",
};

void cpuid_init(void)
{
	char desc[13];
	int eax,ebx,ecx,edx;
	int i, cpuid_level;

	/* Get cpuid level and manufacturer string */
	cpuid(0, &eax, &ebx, &ecx, &edx);
	cpuid_level=eax;
	memcpy(desc, &ebx, 4);
	memcpy(desc+4, &edx, 4);
	memcpy(desc+8, &ecx, 4);
	desc[12]=0;

	/* Get features */
	cpuid(1, &eax, &ebx, &ecx, &edx);
	__this_cpu->features=edx;

	printk("cpu: %u.%uMHz %s cpuid=%i ",
		((__this_cpu->loops_ms*10)/5000),
		((__this_cpu->loops_ms*10)/50)%100,
		desc, cpuid_level);

	for(i=0; i<32; i++) {
		if ( edx & (1<<i) ) printk("%s ", cpuid_str[i]);
	}
	printk("\n");
}

/* Identity map 4MB of memory, and also map it to PAGE_OFFSET */
/* TODO: Don't bother with pagetable if we have PSE */
void setup_initmem(void)
{
	int i;
	pgd_t dir=(pgd_t)INIT_PGD;
	pgt_t tbl=(pgt_t)INIT_PGT;

	for (i=0; i<NR_PDE; i++) dir[i]=0;
	for (i=0; i<NR_PTE; i++) {
		tbl[i]=(i<<PAGE_SHIFT) | PTE_PRESENT | PTE_RW;
	}

	/* Setup paging */
	dir[0]=INIT_PGT | PDE_PRESENT | PDE_RW;
	dir[dir(PAGE_OFFSET)]=INIT_PGT | PDE_PRESENT | PDE_RW;
	load_pdbr(dir);
}

/* Checks the multiboot structures and take whatever info we need */
void multiboot_check(u_int32_t magic, u_int32_t addr)
{
	multiboot_info_t *mbi=(multiboot_info_t *)addr;

	/* Am I booted by a Multiboot-compliant boot loader?  */
	if (magic != MULTIBOOT_BOOTLOADER_MAGIC) {
		printk("PANIC: Invalid magic number: 0x%x\n", (unsigned) magic);
		idle_task();
	}

	if ( mbi->flags&MBF_AOUT && mbi->flags&MBF_ELF ) {
		printk("PANIC: Kernel can't be ELF AND AOUT!\n");
		idle_task();
	}

	if ( !(mbi->flags&MBF_MEM) ) {
		printk("PANIC: ScaraOS relies on bootloader counting memory\n");
		idle_task();
	}

	if ( mbi->flags & MBF_CMDLINE ) {
		cmdline=__va(mbi->cmdline);
	}

	/* BIOS memory map */
	if ( mbi->flags & MBF_MMAP ) {
		p_memory_map addr=(p_memory_map)mbi->mmap;
		while((void *)addr < (void *)(mbi->mmap + mbi->mmap_length)) {
			printk("MEMMAP: 0x%x -> 0x%x (0x%x)\n",
				addr->base_addr_low,
				addr->base_addr_low+addr->length_low,
				addr->type);
			addr = (p_memory_map)
				((char *)addr + addr->size + sizeof (addr->size));
		}
	}

	mem_lo=mbi->mem_lower;
	mem_hi=mbi->mem_upper;
}

/* Map in all physical memory from 4MB onwards */
void map_ram(u_int32_t tot_mem)
{
	u_int32_t bmem=(mem_lo/4)<<PAGE_SHIFT;
	u_int32_t num_pgt=(tot_mem/(1024<<PAGE_SHIFT));
	u_int32_t i,j;

	for(i=1; i<num_pgt; i++) {
		pgt_t tbl=(pgt_t)__bootmem_end;
		pgd_t d=(pgd_t)INIT_PGD;
		u_int32_t addr=(1024*i)<<PAGE_SHIFT;

		/* TODO: Jump to after kernel */
		if ( __bootmem_end>=bmem ) {
			printk("PANIC: Not enough room for page tables!\n");
			idle_task();
		}

		/* First the table entries */
		for (j=0; j<NR_PTE; j++) {
			tbl[j]=((addr++)<<PAGE_SHIFT) | PTE_PRESENT | PTE_RW;
		}

		/* Then the directory entry */
		d[dir((addr+PAGE_OFFSET))]=(u_int32_t)tbl | PDE_PRESENT | PDE_RW;

		__bootmem_end+=PAGE_SIZE;
	}
}

/* Initialise some very low-level memory management stuff such
 * as the page frame array. That is one struct page for each
 * page frame in the system */
void mm_init(void)
{
	pgd_t dir=(pgd_t)INIT_PGD;
	u_int32_t tot_mem, real_end;
	int i, nr_reserved=0;

	/* Calculate size of physical memory */
	tot_mem=(mem_hi+1024)*1024;
	nr_physpages=tot_mem>>PAGE_SHIFT;
	if ( tot_mem > 0x80000000 ) {
		tot_mem=0x80000000;
		printk("mem: only 1GB of RAM supported\n");
	}

	/* Then map it all in */
	map_ram(tot_mem);

	/* Then unset the old identity mapping */
	dir[0]=0;
	__flush_tlb();

	/* Print some stats */
	printk("mem: ram=%uKB(%uMB) data=%uKB code=%uKB\n",
		tot_mem/1024, tot_mem/(1024*1024),
		((void *)&_end - (void *)&_etext) / 1024,
		((void *)&_etext - (void *)&_begin) / 1024);

	/* Allocate a place to store all the struct pages */
	pfa=(struct page *)&_end;
	real_end=(u_int32_t)&_end;
	real_end+=nr_physpages*sizeof(struct page);

	printk("mem: pfa@%x is %u bytes (%u bytes slack)\n",
		pfa, nr_physpages*sizeof(struct page),
		(nr_physpages*sizeof(struct page))%PAGE_SIZE);

	/* Initialise buddy data structures */
	buddy_init();

	/* TODO: take e820 memory maps in to account */
	for(i=0; i<nr_physpages; i++) {
		u_int32_t phys = i<<PAGE_SHIFT;
		int r=0;

		pfa[i].next=NULL;
		pfa[i].prev=NULL;
		pfa[i].count=1;
		
		if ( phys >= (u_int32_t)__pa(&_begin) &&
			phys < (u_int32_t)__pa(real_end)) {
			/* should cover kernel data and code */
			r=1;
		}else if ( i>=160 && i < 256 ) {
			/* 640k-1M reserved for hardware etc... */
			r=1;
		}else if ( phys < __bootmem_end ) {
			/* bottom page reserved for hardware,
			 * next set of pages for initial stack
			 * and page tables etc... */
			r=1;
		}

		if ( r ) {
			pfa[i].flags=PG_reserved;
			nr_reserved++;
		}else{
			pfa[i].flags=0;
			free_pages(page_address(&pfa[i]), 0);
		}
	}
	printk("mem: %u/%u pageframes free (%u reserved)\n",
		nr_freepages, nr_physpages, nr_reserved);
}

/* Init task - the job of this task is to initialise all
 * installed drivers, mount the root filesystem and
 * bootstrap the system */
void init_task(void)
{
	unsigned long *fptr;
	for(fptr=(unsigned long *)&__initcall_start;
		fptr<(unsigned long *)&__initcall_end; fptr++) {
		initcall_t fn=(initcall_t)*fptr;
		fn();
	}

	/* Mount the root filesystem etc.. */
	vfs_init();

	printk("ScaraOS bootup completed\n");
	idle_task();
}

/* Entry point in to the kernel proper */
void setup(void)
{
	struct task *i;

	/* Need this quite quickly */
	idt_init();

	/* Prints out CPU MHz */
	calibrate_delay_loop();

	/* print a pretty message */
	printk("ScaraOS v0.0.2 for IA-32\n");
	if ( cmdline ) printk("cmd: %s\n", cmdline);

	/* Identify CPU features */
	cpuid_init();

	/* Fire up the kernel memory allocators */
	mm_init();

	/* enable hardware interrupts */
	pic_init();

	/* start jiffie counter */
	pit_start_timer1();

	/* setup the idle task */
	sched_init();

	/* Setup the init task */
	i=alloc_page();
	i->pid=1;
	i->t.eip=(u_int32_t)init_task;
	i->t.esp=(u_int32_t)i;
	i->t.esp+=PAGE_SIZE;
	task_to_runq(i);
	sched();

	/* Finally, enable interupts */
	printk("starting idle task...\n");
	sti();
}

void idle_task(void)
{
	asm volatile(
		"idle:\n"
		"rep; nop\n"
		"jmp idle\n");
}
